__all__ = ['Graph','Mode','Inference','Sampler']

import ctypes
import time
import numpy
import struct
from sys import platform
from .functions import *

class Colors:
    INIT  = '\033[11m'
    TRAIN = '\033[10m'
    OTHER = '\033[94m'
    ENDC  = '\033[0m'
    BOLD  = '\033[1m'
    UNDERLINE = '\033[4m'

from pathlib import Path
path = Path(__file__)

if platform == "linux" or platform == "linux2":
	pxpath = str(path.parent / "lib/libpx.so")
elif platform == "darwin":
#	pxpath = path.parent / "lib/libpx.dylib"
	raise TypeError('macOS will be supported soon!')
elif platform == "win32":
	pxpath = str(path.parent / "lib/libpx.dll")

example_model_filename = str(path.parent / "data/5_14.mod")
example_data_filename  = str(path.parent / "data/sin44")

class disc_t(ctypes.Structure):
	_fields_ = [("num_intervals", ctypes.c_uint64),("num_moments", ctypes.c_uint64),("_intervals", ctypes.POINTER(ctypes.c_double)),("_moments", ctypes.POINTER(ctypes.c_double))]

	@property
	def intervals(self):
		return numpy.ctypeslib.as_array(self._intervals,shape=(self.num_intervals,2))

	@property
	def moments(self):
		return numpy.ctypeslib.as_array(self._moments,shape=(self.num_moments,4))

lib = ctypes.cdll.LoadLibrary(pxpath)

lib.create_ctx.restype = ctypes.c_uint64

lib.version.restype = ctypes.c_uint64

lib.ctx_set_code.restype = ctypes.c_bool

lib.discretize.restype = disc_t
lib.discretize.argtypes = [ctypes.c_uint64,ctypes.c_uint64,ctypes.c_uint64,ctypes.c_uint64]

lib.reset_ctx.restype  = ctypes.c_bool
lib.reset_ctx.argtypes = [ctypes.c_uint64]

lib.run_ctx.restype  = ctypes.c_bool
lib.run_ctx.argtypes = [ctypes.c_uint64]

lib.ctx_write_reg.restype  = ctypes.c_bool
lib.ctx_write_reg.argtypes = [ctypes.c_uint64, ctypes.c_char_p, ctypes.c_uint64]

lib.ctx_read_reg.restype  = ctypes.c_uint64
lib.ctx_read_reg.argtypes = [ctypes.c_uint64, ctypes.c_char_p]

class progress_t(ctypes.Structure):
	_fields_ = [("_obj", ctypes.c_double), ("stepsize", ctypes.c_double), ("lam", ctypes.c_double), ("iteration", ctypes.c_uint64), ("max_iterations", ctypes.c_uint64), ("dim", ctypes.c_uint64), ("_w", ctypes.POINTER(ctypes.c_double)), ("_g", ctypes.POINTER(ctypes.c_double)), ("_e", ctypes.POINTER(ctypes.c_double)), ("is_int", ctypes.c_bool), ("_best_obj", ctypes.c_double), ("_best_w", ctypes.POINTER(ctypes.c_double)), ("value_bytes", ctypes.c_uint64)]

	@property
	def obj(self):
		if self.is_int != 0:
			return self._best_obj / 255.0
		else:
			return self._best_obj

	@property
	def weights(self):
		if self.is_int != 0:
			return numpy.ctypeslib.as_array(self._w,shape=(self.dim,)).view('uint64')
		else:
			return numpy.ctypeslib.as_array(self._w,shape=(self.dim,))

	@property
	def weights_extrapolation(self):
		if self.is_int != 0:
			return numpy.ctypeslib.as_array(self._e,shape=(self.dim,)).view('uint64')
		else:
			return numpy.ctypeslib.as_array(self._e,shape=(self.dim,))

	@property
	def best_weights(self):
		if self.is_int != 0:
			return numpy.ctypeslib.as_array(self._best_w,shape=(self.dim,)).view('uint64')
		else:
			return numpy.ctypeslib.as_array(self._best_w,shape=(self.dim,))

	@property
	def gradient(self):
		if self.is_int != 0:
			return numpy.ctypeslib.as_array(self._g,shape=(self.dim,)).view('uint64')
		else:
			return numpy.ctypeslib.as_array(self._g,shape=(self.dim,))

if platform == "win32":
	opt_func = ctypes.WINFUNCTYPE(None, ctypes.POINTER(progress_t))
	prg_func = ctypes.WINFUNCTYPE(None, ctypes.c_uint64, ctypes.c_uint64, ctypes.c_char_p)
else:
	opt_func = ctypes.CFUNCTYPE(None, ctypes.POINTER(progress_t))
	prg_func = ctypes.CFUNCTYPE(None, ctypes.c_uint64, ctypes.c_uint64, ctypes.c_char_p)

class graph_t(ctypes.Structure):
	_fields_ = [("__unused", ctypes.c_uint64),("itype", ctypes.c_uint8),("nodes", ctypes.c_uint64),("edges", ctypes.c_uint64),("A", ctypes.POINTER(ctypes.c_uint64))]

	@property
	def edgelist(self):
		res = numpy.ctypeslib.as_array(self.A, shape=(self.edges,2))
		res.flags.writeable = False
		return res

class st_graph_t(ctypes.Structure):
	_fields_ = [("__unused", ctypes.c_uint64),("itype", ctypes.c_uint8),("T", ctypes.c_uint64),("base", ctypes.POINTER(graph_t)),("Tm1inv", ctypes.c_float)]

	@property
	def nodes(self):
		H = self.base.contents
		return H.nodes * self.T

	@property
	def edges(self):
		H = self.base.contents
		return H.edges * self.T + (self.T-1) * ( H.nodes + 2 * H.edges )

	@property
	def edgelist(self):
		if not hasattr(self, 'E'):
			self.E = numpy.zeros((self.edges,2), dtype=numpy.uint64)
			H = self.base.contents

			for v in range(H.nodes):
				for t in range(self.T-1):
					e = v*(self.T-1)+t
					self.E[e][0] = t*H.nodes+v
					self.E[e][1] = (t+1)*H.nodes+v

			for f in range(H.edges):
				a = H.edgelist[f][0]
				b = H.edgelist[f][1]
				for t in range(self.T-1):
					e = (self.T-1)*H.nodes + f*3*(self.T-1)+t*3
					self.E[e+0][0] = t*H.nodes+a
					self.E[e+0][1] = t*H.nodes+b
					self.E[e+1][0] = t*H.nodes+a
					self.E[e+1][1] = (t+1)*H.nodes+b
					self.E[e+2][0] = (t+1)*H.nodes+a
					self.E[e+2][1] = t*H.nodes+b

				e =  (self.T-1)*H.nodes + (self.T-1)*3*H.edges + f
				self.E[e][0] = (self.T-1)*H.nodes+a
				self.E[e][1] = (self.T-1)*H.nodes+b

			self.E.flags.writeable = False

		return self.E

	def spatial_vertex(self, v):
		H = self.base.contents
		return v % H.nodes

	def time(self, v):
		H = self.base.contents
		return (v-self.spatial_vertex(v)) / H.nodes

	def is_spatial_edge(self, e):
		return self.time(self.edgelist[e][0]) == self.time(self.edgelist[e][1])
		

class model_t(ctypes.Structure):
	_fields_ = [("itype", ctypes.c_uint8),("vtype", ctypes.c_uint8),("from_file", ctypes.c_bool),("G", ctypes.c_uint64),("H", ctypes.c_uint64),("w", ctypes.POINTER(ctypes.c_double)),("E", ctypes.POINTER(ctypes.c_double)),("Y", ctypes.POINTER(ctypes.c_uint64)),("Ynames", ctypes.POINTER(ctypes.c_uint64)),("Xnames", ctypes.POINTER(ctypes.c_uint64)),("dimension", ctypes.c_uint64),("gtype", ctypes.c_uint64),("T", ctypes.c_uint64),("reparam", ctypes.c_uint64),("K", ctypes.c_uint64),("num_instances", ctypes.c_uint64),("llist", ctypes.c_uint64),("clist", ctypes.c_uint64)]

	def __init__(self):
		self.woff = []

	def init(self):
		offset = 0
		self.woff = []
		for e in range(self.graph.edges):
			L = self.states[self.graph.edgelist[e][0]] * self.states[self.graph.edgelist[e][1]]
			self.woff.append(offset)
			offset = offset + L
		self.woff.append(offset)

	@property
	def dim(self):
		return self.dimension

	@property
	def time_frames(self):
		return self.T

	@property
	def weights(self):
		if self.vtype == 3:
			return numpy.ctypeslib.as_array(self.w,shape=(self.dimension,)).view('uint64')
		else:
			return numpy.ctypeslib.as_array(self.w,shape=(self.dimension,))

	def slice_edge(self,e,A):
		if not isinstance(A, numpy.ndarray):
			raise TypeError('A must be of type numpy.ndarray')

		if len(A) != self.dim:
			raise TypeError('A must be ' + str(self.dim) + ' dimensional')

		return A[int(self.woff[e]):int(self.woff[e+1])]

	def slice_edge_state(self,e,x,y,A):
		w = self.slice_edge(e,A)

		s = self.graph.edgelist[e][0]
		t = self.graph.edgelist[e][1]

		return w[x*self.states[t]+y]

	@property
	def statistics(self):
		res = 0
		if self.vtype == 3:
			res = res = numpy.ctypeslib.as_array(self.E,shape=(self.dimension,)).view('uint64') / self.num_instances
		else:
			res = res = numpy.ctypeslib.as_array(self.E,shape=(self.dimension,)) / self.num_instances
		res.flags.writeable = False
		return res

	def phi(self, x):
		if not isinstance(x, numpy.ndarray):
			raise TypeError('x must be of type numpy.ndarray')

		if len(x) != self.graph.nodes:
			raise TypeError('x must be ' + str(self.graph.nodes) + ' dimensional')

		phi_x = numpy.zeros(shape=(self.dim,))

		for e in range(self.graph.edges):

			s = self.graph.edgelist[e][0]
			t = self.graph.edgelist[e][1]

			if x[s]>=self.states[s] or x[t]>=self.states[t]:
				TypeError('Some values of x exceed the state space')

			y = int(self.woff[e] +  x[s] * self.states[t] + x[t])

			phi_x[y] = 1

		return phi_x

	def save(self, filename):
		write_register("OVW",1)
		L=[]
		L.append("MFN \""+filename+"\";")
		L.append("STORE MPT;")
		recode(L)
		run()

	def predict(self, x, generic_progress_hook=0, iterations=100):
		if not isinstance(x, numpy.ndarray):
			raise TypeError('x must be of type numpy.ndarray')

		data_ptr = ctypes.c_uint64(x.ctypes.data)

		if generic_progress_hook != 0:
			f3 = prg_func(generic_progress_hook)
			write_register("CBP",ctypes.c_uint64.from_buffer(f3).value)
		else:
			write_register("CBP",0)

		write_register("PGX",0)
		write_register("MIL",iterations)
		write_register("INF",int(Inference.belief_propagation))

		L=[]
		L.append("EDP "+str(data_ptr.value)+";")
		L.append("NXX "+str(len(x))+";")
		L.append("GPX "+str(len(x)*len(x[0])*2)+";")
		L.append("LDX DPT;")
		L.append("PREDICT;")
		
		recode(L)
		run()

	def sample(self, sampler=Sampler.apx_perturb_and_map, generic_progress_hook=0, num_samples=1, iterations=100, perturbation=0.1, burn=10, resamplings=100):
		if not isinstance(sampler, Sampler):
			raise TypeError('sampler must be an instance of Sampler Enum')

		x = -numpy.ones(shape=(num_samples, self.graph.nodes), dtype=numpy.uint16)
		data_ptr = ctypes.c_uint64(x.ctypes.data)

		L=[]
		L.append("EDP "+str(data_ptr.value)+";")
		L.append("NXX "+str(len(x))+";")
		L.append("GPX "+str(len(x)*len(x[0])*2)+";")

		if generic_progress_hook != 0:
			f3 = prg_func(generic_progress_hook)
			write_register("CBP",ctypes.c_uint64.from_buffer(f3).value)
		else:
			write_register("CBP",0)

		if sampler == Sampler.apx_perturb_and_map:
			write_register("PAM",integer_from_float(perturbation))
			write_register("MIL",iterations)
			write_register("INF",int(Inference.belief_propagation))

		elif sampler == Sampler.gibbs:
			write_register("PAM",0)
			write_register("GBR",burn)
			write_register("GRE",resamplings)

		L.append("LDX DPT;")
		L.append("SAMPLE;")
		recode(L)
		run()

		return x

	@property
	def graph(self):
		if not hasattr(self, 'LOCAL_G'):
			if self.gtype == 11:
				self.LOCAL_G = ctypes.cast(self.G,ctypes.POINTER(st_graph_t)).contents
			else:
				self.LOCAL_G = ctypes.cast(self.G,ctypes.POINTER(graph_t)).contents
		return self.LOCAL_G


	@property
	def states(self):
		res = numpy.ctypeslib.as_array(self.Y,shape=(self.graph.nodes,))
		return res

	def infer(self, inference = Inference.belief_propagation, iterations=100, k=3):
		if not isinstance(inference, Inference):
			raise TypeError('inference must be an instance of Inference Enum')

		write_register("KXX",k)
		write_register("MIL",iterations)

		write_register("INF",int(inference))
		L=[]
		L.append("INFER;")
		recode(L)
		run()

		P = ctypes.cast(int(read_register("PPT")), ctypes.POINTER(ctypes.c_double))

		res = numpy.ctypeslib.as_array(P,shape=(self.dimension,))
		res.flags.writeable = False
		return res, float_from_integer(read_register("LNZ"))

	@property
	def addr(self):
		return self

ctx = ctypes.c_uint64(lib.create_ctx())

def write_register(name,val):
	l = len(name)
	buff = ctypes.create_string_buffer(l+1)
	buff.value = name.encode('utf-8')
	ptr = (ctypes.c_char_p)(ctypes.addressof(buff))
	return lib.ctx_write_reg(ctx,ptr,val)

write_register("SEED",int(round(time.time() * 1000)))

def version():
	return lib.version()

def discretize(data, num_states, targets=None):
	if not isinstance(data, numpy.ndarray):
		raise TypeError('data must be of type numpy.ndarray')

	R = numpy.zeros(shape=(len(data),len(data[0])), dtype=numpy.uint16)
	M = []

	if targets == None:
		t = range(len(data[0]))
	else:
		t = targets

	for col in t:
		col_data = numpy.ascontiguousarray(data[:,col])
		distinct = len(numpy.unique(col_data))
		if distinct > num_states:
			result = numpy.zeros(shape=(len(data),), dtype=numpy.uint16)
			disc_info = lib.discretize(ctypes.c_uint64(result.ctypes.data), ctypes.c_uint64(col_data.ctypes.data), ctypes.c_uint64(len(data)), ctypes.c_uint64(num_states))
			M.append(disc_info)
			R[:,col] = result
		else:
			M.append(None)
			R[:,col] = col_data
	return R,M

def undiscretize(data, M):
	if not isinstance(data, numpy.ndarray):
		raise TypeError('data must be of type numpy.ndarray')

	R = numpy.zeros(shape=(len(data),len(data[0])))

	with numpy.nditer(data, flags=['multi_index']) as it:
		while not it.finished:
			row = it.multi_index[0]
			col = it.multi_index[1]
			R[row,col] = M[col].moments[it[0]][0]
			it.iternext()
	return R

def run():
	if ctx is None:
		assert False
	lib.run_ctx(ctx)

def read_register(name):
	l = len(name)
	buff = ctypes.create_string_buffer(l+1)
	buff.value = name.encode('utf-8')
	ptr = (ctypes.c_char_p)(ctypes.addressof(buff))
	return lib.ctx_read_reg(ctx,ptr)

def recode(code):
	n = len(code)
	l = 0
	for stmt in code:
		if len(stmt) > l:
			l = len(stmt)
	buffs = [ctypes.create_string_buffer(l+1) for i in range(len(code))]
	for index, stmt in enumerate(code):
		buffs[index].value = code[index].encode('utf-8')
	ptrs = (ctypes.c_char_p*n)(*map(ctypes.addressof, buffs))
	lib.ctx_set_code(ctx,ptrs,n)
	return n

def set_seed(s):
	write_register("SEED",s)

def float_from_integer(val):
    return struct.unpack('d', struct.pack('N', val))[0]

def integer_from_float(val):
    return struct.unpack('N', struct.pack('d', val))[0]

def KL(p,q):
	if not isinstance(p, numpy.ndarray):
		raise TypeError('p must be of type numpy.ndarray')

	if not isinstance(q, numpy.ndarray):
		raise TypeError('q must be of type numpy.ndarray')

	if len(p) != len(q):
		raise TypeError('p and q must have same length')

	res = 0
	for i in range(len(p)):
		res = res + p[i] * numpy.log(p[i]/q[i])

	return res

def load_model(filename):
	L=[]
	L.append("MFN \""+filename+"\";")
	L.append("LDX MPT;")
	recode(L)
	run()
	model = ctypes.cast(read_register("MPT"),ctypes.POINTER(model_t)).contents
	model.init()
	return model

def train(data, graph, mode, inference = Inference.belief_propagation, iters=1000, seed=0, k=3, cont=False, model=0, initial_stepsize=0.1, opt_progress_hook=0, generic_progress_hook=0, opt_regularization_hook=0, opt_proximal_hook=0, T=1, edge_fraction=0.1, lam=0):
	if not isinstance(mode, Mode):
		raise TypeError('mode must be an instance of Mode Enum')

	if not (isinstance(graph, Graph) or isinstance(graph, numpy.ndarray)):
		raise TypeError('graph must be an instance of Graph Enum or numpy.ndarray (edgelist)')
	if isinstance(graph, numpy.ndarray) and len(graph[0])!=2:
		raise TypeError('graph must be an edgelist')

	if not isinstance(inference, Inference):
		raise TypeError('inference must be an instance of Inference Enum')

	if not isinstance(data, numpy.ndarray):
		raise TypeError('data must be of type numpy.ndarray')

	if cont == True and model == 0:
		raise ValueError('model must be provided when cont is True')

	if model!=0 and not isinstance(model, model_t):
		raise TypeError('model must be an instance of model_t Class')

	lib.reset_ctx(ctx)

	if opt_progress_hook != 0:
		f1 = opt_func(opt_progress_hook)
		write_register("CBO",ctypes.c_uint64.from_buffer(f1).value)
	else:
		write_register("CBO",0)

	if opt_regularization_hook != 0:
		f2 = opt_func(opt_regularization_hook)
		write_register("CBU",ctypes.c_uint64.from_buffer(f2).value)
	else:
		write_register("CBU",0)

	if generic_progress_hook != 0:
		f3 = prg_func(generic_progress_hook)
		write_register("CBP",ctypes.c_uint64.from_buffer(f3).value)
	else:
		write_register("CBP",0)

	if opt_proximal_hook != 0:
		f4 = opt_func(opt_proximal_hook)
		write_register("CPR",ctypes.c_uint64.from_buffer(f4).value)
	else:
		write_register("CPR",0)

	L=[]
	L.append("DEL MPT;");

	if mode == Mode.integer:
		L.append("idx_t UINT64")
		L.append("val_t UINT64")
		L.append("ALG IGD")
	else:
		write_register("STP",integer_from_float(initial_stepsize))
		L.append("idx_t UINT64")
		L.append("val_t FLT64")
		L.append("ALG FL1")

	write_register("KXX",k)
	if seed != 0:
		write_register("SEED",seed)
	write_register("MIS",ord('?'))
	write_register("SEP",ord(','))
	write_register("OVW",1)
	write_register("HED",0)
	write_register("LSN",2)

	write_register("INF",int(inference))
	write_register("MIL",100)
	write_register("MIO",iters)
	write_register("EPO",0)
	write_register("LAM",integer_from_float(lam))
	write_register("ELAM",0)

	if mode >= Mode.strf_linear and mode <= Mode.strf_inv_exponential:
		write_register("TXX",T)
		write_register("YYC",1)
	else:
		write_register("TXX",1)
		write_register("YYC",0)
		write_register("REP",0)

	if cont == False:
		data_ptr = ctypes.c_uint64(data.ctypes.data)

		L.append("EDP "+str(data_ptr.value)+";")
		L.append("NXX "+str(len(data))+";")
		L.append("GPX "+str(len(data)*len(data[0])*2)+";")
		L.append("LDX DPT;")

		if isinstance(graph, numpy.ndarray):
			M = numpy.amax(graph)
			V = numpy.unique(graph)
			n = len(V)
			m = len(graph)
			s = numpy.sum(V)

			if (n != M+1) or (s != (M*(M+1))/2):
				raise ValueError('please provide a valid edge list')

			data_ptr = ctypes.c_uint64(graph.ctypes.data)
			L.append("EAP "+str(data_ptr.value)+";")
			L.append("GPX "+str(n)+";")
			L.append("RES "+str(m)+";")
			write_register("GRA",int(Graph.file))
		else:
			write_register("GRA",int(graph))

		L.append("LDX GPT;")

		if not isinstance(graph, numpy.ndarray)and graph == Graph.auto:
			L.append("PEL "+str(edge_fraction)+";")
		L.append("CREATE;")

		if mode == Mode.dbt:
			L.append("DEL MPT;")
			L.append("GRA OTHER;")
			L.append("LSN 1024;")
			L.append("BOLTZMANNTREE;")
			L.append("INITLATENT;")
			L.append("CREATE;")

		if mode >= Mode.strf_linear and mode <= Mode.strf_inv_exponential and T>1:
			L.append("DEL MPT;")
			L.append("GRA OTHER;")
			L.append("STGRAPH;")
			L.append("REP "+str(int(mode))+";")
			L.append("CREATE;")

	else:
		write_register("GRA",int(Graph.other))
		write_register("MPT",model.addr)

	if mode != Mode.integer and opt_regularization_hook == 0 and opt_proximal_hook == 0:
		L.append("CLOSEDFORM;")
	L.append("ESTIMATE;")

	recode(L)
	run()

	mod = ctypes.cast(read_register("MPT"),ctypes.POINTER(model_t)).contents
	mod.init()

	res = read_register("RES")
	if mode != Mode.integer:
		res = float_from_integer(res)

	return mod, res
